import datetime

class RetainDocumentPolicy():
    for_release_date(self):
        today = datetime.datetime.now()
        gdpr_implementation_date = datetime.datetime.strptime('25-05-2018', '%d-%m-%Y')
        return gdpr_implementation_date <= release_date <= today



class RootDocumentSpecification():
    def __init__(self, root):
        self.root = root

    def is_satisfied_by(self, document):
        split = self.root.split('/')
        doc_index = split.index(doc) if document in split else -1
        last_index = len(split) - 1
        return (doc_index == last_index)



from enum import Enum

class Color(Enum):
    WHITE = 'white'
    GREY = 'lightgrey'
    BLACK = 'black'

class ReachabilityNode(object):
    def __init__(self, key=None):
        self.key = key
        self.parent = None
        self.children = []
        self.color = Color.WHITE

    def get_color(self):
        return self.color

    def get_children(self):
        return self.children

    def set_color(self, color):
        self.color = color

    def set_parent(self, parent):
        self.parent = parent

    def add_child(self, child):
        child.set_parent(self)
        self.children.append(child)

class GDPRNode(ReachabilityNode):
    def __init__(self, key):
        key = 'GDPR'
        super().__init__(key)

class DPANode(ReachabilityNode):
    def __init__(self, key, dpa):
        # satellite data
        self.dpa = dpa
        super().__init__(key)

class SourceNode(ReachabilityNode):
    def __init__(self, key, source):
        # satellite data
        self.source = source
        super().__init__(key)

class LabelNode(ReachabilityNode):
    def __init__(self, key, label):
        # satellite data
        self.label = label
        super().__init__(key)

class XPathNode(ReachabilityNode):
    def __init__(self, key, xpath):
        # satellite data
        self.xpath = xpath
        super().__init__(key)



import os
import math
import socket
import sys
import time
import requests
import json
import datetime
import hashlib
import dateparser
import requests.exceptions
import urllib3.exceptions

from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import random

class Greece(DPA):
    def __init__(self, path=os.curdir):
        country_code='GR'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path=None):
        source = {
            "host": "https://www.dpa.gr"
            # "start_path": "/portal/page?_pageid=33,43547&_dad=portal&_schema=PORTAL"
            #"start_path": "/el/enimerwtiko/prakseisArxis?field_year_from=2018&field_year_to=&field_category=239&field_thematic=All&field_protocol_number=&field_keywords=&page=0"
        }
        host = source['host']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)

        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Recommendations(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Opinions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guidelines(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Greece Decision Documents ===========================")

        added_docs = []
        dict_hashcode = {}
        pagination = self.update_pagination(start_path='/el/enimerwtiko/prakseisArxis?field_year_from=2018&field_year_to=&field_category=239&field_thematic=All&field_protocol_number=&field_keywords=&page=0')

        # This list stores the pages we have visited
        visited_pages = []
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            tbody = results_soup.find('tbody')
            assert tbody

            iterator = 1
            duplicate_title = 1
            for tr in tbody.find_all('tr'):
                assert tr

                print("\n------------ Document " + str(iterator) + " ------------")
                iterator += 1

                date_tag = tr.find('time', datetime='00Z')
                assert date_tag
                date_str = date_tag.get_text()
                print('\tDocument Date:\t' + date_str)
                document_year = date_str[-4:]

                stop_date = '25/05/2018'
                stop_date = time.strptime(stop_date, "%d/%m/%Y")
                doc_date = time.strptime(date_str, "%d/%m/%Y")
                if doc_date < stop_date:
                    print("\tSkipping outdated document")
                    return added_docs

                a_tag = tr.find('a')
                document_title = a_tag.get_text()
                assert document_title
                print('\tDocument Title:\t' + document_title)

                # Create the document has using the document title
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = a_tag.get('href')
                assert document_href
                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://www.dpa.gr' + document_href

                if to_print:
                    print("\tDocument:\t", document_hash)

                # Get document response object
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=3)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                span_tag = document_soup.find('span', class_='file-link')
                assert span_tag

                span_a_tag = span_tag.find('a')
                assert span_a_tag
                pdf_href = span_a_tag.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://www.dpa.gr' + pdf_href

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url, timeout=10)
                    pdf_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

                # in case several documents share same title
                dpa_folder = self.path
                date_part = date_str.replace('/', '_')
                if document_hash in dict_hashcode:
                    # documents have the same hashcode, but different dates
                    document_hash = document_hash + '-' + date_part
                document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash

                broken_pdf = ["7f3aba140d5ef8ef566db7161d4b46c1", "9d0b0819a3d15cdc006a586f957de592", "50a8044fb0ab8af21402313fe290aaa1"]
                if document_hash in broken_pdf:
                    continue

                # document_folder = dpa_folder + '/greece' + '/' + 'Decisions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(link_text)

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': pdf_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    dict_hashcode[document_hash] = date_part
                    added_docs.append(document_hash)

                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            # Add the next page to the pagination object
            ul = results_soup.find('ul', class_='pagination js-pager__items')
            assert ul
            # The last two li tags contain links for 'next' and 'end' buttons
            li_list = ul.find_all('li')
            assert li_list
            # The second to last element should contain the 'next page' link
            a = li_list[-2].find('a')
            assert a
            a_href = a.get('href')
            assert a_href
            page_link = 'https://www.dpa.gr/el/enimerwtiko/prakseisArxis' + a_href
            # If the index where next page usually is leads to a page that scraper has already visited -> don't add
            # to pagination object
            if page_link in visited_pages:
                print("There seems to be no more pages to look at")
                continue
            else:
                print("\nNext Page: " + page_link)
                visited_pages.append(page_link)
                pagination.add_item(page_link)
        return added_docs

    def get_docs_Recommendations(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Greece Recommendations ===========================")
        added_docs = []
        added_date = []
        pagination = self.update_pagination(start_path='/el/enimerwtiko/prakseisArxis?field_year_from=&field_year_to=&field_category=246&field_thematic=All&field_protocol_number=&field_keywords=')

        # This list stores the pages we have visited
        visited_pages = []
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            tbody = results_soup.find('tbody')
            assert tbody

            iterator = 1
            duplicate_title = 1
            for tr in tbody.find_all('tr'):
                assert tr

                print("\n------------ Document " + str(iterator) + " ------------")
                iterator += 1

                date_tag = tr.find('time', datetime='00Z')
                assert date_tag
                date_str = date_tag.get_text()
                print('\tDocument Date:\t' + date_str)
                document_year = date_str[-4:]

                if int(document_year) < 2018:
                    print("\tSkipping outdated document")
                    continue

                a_tag = tr.find('a')
                document_title = a_tag.get_text()
                assert document_title
                print('\tDocument Title:\t' + document_title)

                # Create the document has using the document title
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = a_tag.get('href')
                assert document_href
                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://www.dpa.gr' + document_href

                if to_print:
                    print("\tDocument:\t", document_hash)

                # Get document response object
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                span_tag = document_soup.find('span', class_='file-link')
                assert span_tag

                span_a_tag = span_tag.find('a')
                assert span_a_tag
                pdf_href = span_a_tag.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://www.dpa.gr' + pdf_href

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url)
                    pdf_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

                dpa_folder = self.path
                # in case several documents share same title
                dpa_folder = self.path
                if document_hash in added_docs and date_str not in added_date:
                    document_folder = dpa_folder + '/' + 'Recommendations' + '/' + document_hash + '_' + str(duplicate_title)
                    duplicate_title += 1
                else:
                    document_folder = dpa_folder + '/' + 'Recommendations' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    # TODO: Ask about self.language_code -> its 'en' right now...
                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(link_text)

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                    added_date.append(date_str)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

        return added_docs

    def get_docs_Opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Greece Opinions ===========================")
        added_docs = []
        added_date = []
        pagination = self.update_pagination(start_path='/el/enimerwtiko/prakseisArxis?field_year_from=&field_year_to=&field_category=238&field_thematic=All&field_protocol_number=&field_keywords=')

        # This list stores the pages we have visited
        visited_pages = []
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            tbody = results_soup.find('tbody')
            assert tbody

            iterator = 1
            duplicate_title = 1
            for tr in tbody.find_all('tr'):
                assert tr

                print("\n------------ Document " + str(iterator) + " ------------")
                iterator += 1

                date_tag = tr.find('time', datetime='00Z')
                assert date_tag
                date_str = date_tag.get_text()
                print('\tDocument Date: ' + date_str)
                document_year = date_str[-4:]

                if int(document_year) < 2018:
                    print("\tSkipping outdated document")
                    continue

                a_tag = tr.find('a')
                document_title = a_tag.get_text()
                assert document_title
                print('\tDocument Title: ' + document_title)

                # Create the document has using the document title
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = a_tag.get('href')
                assert document_href
                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://www.dpa.gr' + document_href

                if to_print:
                    print("\tDocument:\t", document_hash)

                # Get document response object
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=(2, 2))
                    document_response.raise_for_status()
                # Added more errors to catch
                except requests.exceptions.ReadTimeout as error:
                    if to_print:
                        print(error)
                    pass
                except requests.exceptions.ConnectionError as error:
                    if to_print:
                        print(error)
                    pass
                except urllib3.exceptions.ReadTimeoutError as error:
                    if to_print:
                        print(error)
                    pass
                except socket.timeout as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                span_tag = document_soup.find('span', class_='file-link')
                assert span_tag

                span_a_tag = span_tag.find('a')
                assert span_a_tag
                pdf_href = span_a_tag.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://www.dpa.gr' + pdf_href

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url, timeout=(2, 2))
                    pdf_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                except requests.exceptions.ConnectionError as error:
                    if to_print:
                        print(error)
                    pass
                except urllib3.exceptions.ReadTimeoutError as error:
                    if to_print:
                        print(error)
                    pass
                except socket.timeout as error:
                    if to_print:
                        print(error)
                    pass

                if pdf_response is None:
                    continue

                dpa_folder = self.path
                if document_hash in added_docs and date_str not in added_date:
                    document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash + '_' + str(duplicate_title)
                    duplicate_title += 1
                else:
                    document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)
                    # If a pdf fails to convert to text (because pdf is broken), print error
                    try:
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(link_text)
                    except:
                        print("Failed to convert pdf to text document.")
                        pass

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                    added_date.append(date_str)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            # Add the next page to the pagination object
            ul = results_soup.find('ul', class_='pagination js-pager__items')
            assert ul
            # The last two li tags contain links for 'next' and 'end' buttons
            li_list = ul.find_all('li')
            assert li_list
            # The second to last element should contain the 'next page' link
            a = li_list[-2].find('a')
            assert a
            a_href = a.get('href')
            assert a_href
            page_link = 'https://www.dpa.gr/el/enimerwtiko/prakseisArxis' + a_href
            # If the index where next page usually is leads to a page that scraper has already visited -> don't add
            # to pagination object
            if page_link in visited_pages:
                print("There seems to be no more pages to look at")
                continue
            else:
                print("\nNext Page: " + page_link)
                visited_pages.append(page_link)
                pagination.add_item(page_link)
        return added_docs

    def get_docs_Guidelines(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Greece Guidelines ===========================")
        added_docs = []
        added_date = []
        pagination = self.update_pagination(start_path='/el/enimerwtiko/prakseisArxis?field_year_from=&field_year_to=&field_category=245&field_thematic=All&field_protocol_number=&field_keywords=')

        # This list stores the pages we have visited
        visited_pages = []
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            tbody = results_soup.find('tbody')
            assert tbody

            iterator = 1
            duplicate_title = 1
            for tr in tbody.find_all('tr'):
                assert tr

                print("\n------------ Document " + str(iterator) + " ------------")
                iterator += 1

                date_tag = tr.find('time', datetime='00Z')
                assert date_tag
                date_str = date_tag.get_text()
                print('\tDocument Date:\t' + date_str)
                document_year = date_str[-4:]

                if int(document_year) < 2018:
                    print("\tSkipping outdated document")
                    continue

                a_tag = tr.find('a')
                document_title = a_tag.get_text()
                assert document_title
                print('\tDocument Title:\t' + document_title)

                # Create the document has using the document title
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = a_tag.get('href')
                assert document_href
                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://www.dpa.gr' + document_href

                if to_print:
                    print("\tDocument:\t", document_hash)

                # Get document response object
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                span_tag = document_soup.find('span', class_='file-link')
                assert span_tag

                span_a_tag = span_tag.find('a')
                assert span_a_tag
                pdf_href = span_a_tag.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://www.dpa.gr' + pdf_href

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url)
                    pdf_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

                dpa_folder = self.path
                if document_hash in added_docs and date_str not in added_date:
                    document_folder = dpa_folder + '/' + 'Guidelines' + '/' + document_hash + '_' + str(duplicate_title)
                    duplicate_title += 1
                else:
                    document_folder = dpa_folder + '/' + 'Guidelines' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(link_text)

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                    added_date.append(date_str)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

        return added_docs

    # Only visits first page because further pages contain only outdated documents
    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Greece Annual Reports ===========================")
        added_docs = []
        pagination = self.update_pagination(start_path='/enimerwtiko/etisies-ektheseis')

        # This list stores the pages we have visited
        visited_pages = []
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            region = results_soup.find('div', class_='region region-content')
            assert region

            view_content = region.find('div', class_='view-content')
            assert view_content

            clearfix = view_content.find('div', class_='views-col')
            assert clearfix

            iterator = 1
            # Only look at outer 'div' tags
            for div in clearfix.find_all('div', class_='events-teaser'):
                assert div

                print("\n------------ Document " + str(iterator) + " ------------")
                iterator += 1

                a_tag = div.find('a')
                assert a_tag

                document_year = a_tag.get_text()[-4:]

                print('\tDocument Date:\t' + document_year)

                if int(document_year) < 2018:
                    print("\tSkipping outdated document")
                    continue

                document_title = a_tag.get_text()
                assert document_title
                print('\tDocument Title:\t' + document_title)

                # Create the document has using the document title
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = a_tag.get('href')
                assert document_href
                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://www.dpa.gr' + document_href

                if to_print:
                    print("\tDocument:\t", document_hash)

                # Get document response object
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=5)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                article = document_soup.find('article', role='article')
                assert article

                content = article.find('div', class_='content')
                assert content

                pdf_a = content.find('a')
                assert pdf_a

                pdf_href = pdf_a.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://www.dpa.gr' + pdf_href

                print("\tPDF URL: " + pdf_url)

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url, timeout=5)
                    pdf_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash
                # document_folder = dpa_folder + '/greece' + '/' + 'Annual Reports' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)

                    try:
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(link_text)
                    except:
                        print("Failed to convert pdf to text document")
                        pass

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_year,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

        return added_docs




import os
import math
import requests
import json
import datetime
import hashlib
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy

class Estonia(DPA):
    def __init__(self, path=os.curdir):
        country_code='EE'
        super().__init__(country_code, path)


    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Prescriptions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Instructions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Prescriptions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Estonia Prescriptions ===========================")
        iteration = 1
        existed_docs = []
        dict_hashcode = {}
        source = {
            "host": "https://www.aki.ee",
            "start_path": "/et/inspektsioon-kontaktid/menetlusotsused/ettekirjutused"
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print('page not exist')
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        region_inner = results_soup.find('div', class_='region-sidebar-first-inner')
        menu = region_inner.find('ul', class_='menu')
        for li in menu.find_all('li'):
            result_href = li.find('a').get('href')
            result_link = host + result_href
            result_text = li.find('a').get_text()
            result_year = result_text.split()[-1]
            if result_year < '2018':
                continue
            result_source = self.get_source(page_url=result_link)
            if result_source is None:
                continue
            pages_soup = BeautifulSoup(result_source.text, 'html.parser')
            assert pages_soup
            region_content_inner = pages_soup.find('div', class_='region-content-inner')
            block_system = region_content_inner.find('div', class_='block-system')

            # get year
            page_title = region_content_inner.find(id="page-title")
            page_title = page_title.get_text()
            document_year = page_title.split()[1]

            field_item_even = block_system.find('div', class_='field-item even')
            for p in field_item_even.find_all('p'):

                document = p.find('a')
                if document is None:
                    continue

                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                document_href = document.get('href')
                if document_href.startswith('https'):
                    document_url = document_href
                else:
                    document_url = host + document_href
                if document_href.endswith('pdf') is False:
                    continue
                document_title = document.get_text()
                print('document_title: ', document_title)
                print('\tdocument_url: ', document_url)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                if document_hash in dict_hashcode and dict_hashcode[document_hash] == date:
                    print('\tSkipping existing document:\t', document_hash)

                # get the date from title or href,
                # documents before 2023 have date, while documents after 2023 don't have date
                # Set default date as the document year
                date = document_year
                if document_year == '2023':
                    date = "2023"
                else:
                    for i in range(len(document_title)):
                        if document_title[i].isdigit():
                            date_str = document_title[i: i+10]
                            break
                    try:
                        tmp = datetime.datetime.strptime(date_str, '%d.%m.%Y')
                        date = datetime.date(tmp.year, tmp.month, tmp.day)
                        if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                            continue
                        date = date.strftime('%d/%m/%Y')
                    except ValueError:
                        print("\tnot format date in document title, use default date")
                        pass
                print('\tdocument_date: ', date)
                print('\tdocument_hash: ', document_hash)

                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass

                # Bad pdf file, ignore it
                if document_hash == "6bfd730bfc5d592ba92af457299a9358":
                    continue

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Prescriptions' + '/' + document_hash
                # document_folder = dpa_folder + '/estonia' + '/' + 'Prescriptions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    if document_response is None:
                        continue

                    document_content = document_response.content
                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_content)

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        try:
                            document_text = PDFToTextService().text_from_pdf_path(
                                document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                        except FileNotFoundError:
                            print("can't convert pdf to txt file")

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                    dict_hashcode[document_hash] = date
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

        return existed_docs

    def get_docs_Instructions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Estonia Instructions ===========================")
        existed_docs = []
        source = {
            "host": "https://www.aki.ee",
            "start_path": "/et/koik-juhised-loetelus"
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print('page not exist')
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        block_main = results_soup.find('div', class_='block-system-main')
        content_clearfix = block_main.find('div', class_='content clearfix')
        # type 1 files
        field_name_body = content_clearfix.find('div', class_='field-name-body')
        field_item = field_name_body.find('div', class_='field-item even')
        tbody = field_item.find('tbody')
        for tr in tbody.find_all('tr'):

            year_list = []
            for td in tr.find_all('td'):

                if td.get_text() == 'Teiste asutustega koostöös loodud juhendid':
                    break
                # find the year of document
                if td.find('a') == None:
                    candidate_year = td.get_text()
                    if candidate_year.isdigit():
                        year_list.append(candidate_year)
                    continue
                document_href = td.find('a').get('href')
                document_title = td.find('a').get_text()
            if len(year_list) == 0:
                continue
            year = year_list[-1]
            if year < '2018':
                continue
            print('document_title:', document_title)
            if document_href.startswith('https'):
                document_url = document_href
            else:
                document_url = host + document_href
            print('\tdocument_url:', document_url)
            print('\tyear: ',year)
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            print('\tdocument_hash: ', document_hash)
            if document_hash in existing_docs and overwrite == False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Instructions' + '/' + document_hash
            # document_folder = dpa_folder + '/estonia' + '/' + 'Instructions' + '/' + document_hash
            try:
                os.makedirs(document_folder)
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_content = document_response.content
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_content)
                with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                    document_text = PDFToTextService().text_from_pdf_path(
                        document_folder + '/' + self.language_code + '.pdf')
                    f.write(document_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': year,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        # type 2 files
        field_name_field = content_clearfix.find('div', class_='field-name-field-files')
        field_item = field_name_field.find('div', class_='field-item even')
        tbody = field_item.find('tbody')
        for tr in tbody.find_all('tr', class_='odd'):

            year_str = tr.find('td', class_='extended-file-field-table-date').get_text()
            document_section = tr.find('td', class_='extended-file-field-table-filename')
            document_url = document_section.find('a').get('href')
            document_title = document_section.find('a').get_text()
            year = year_str.split()[0].split('.')[-1]
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if year < '2018':
                continue
            print('document_title:', document_title)
            print('\tdocument_url:', document_url)
            print('\tdocument_hash: ',document_hash)

            if document_hash in existing_docs and overwrite == False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            existed_docs.append(document_hash)
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Instructions' + '/' + document_hash
            # document_folder = dpa_folder + '/estonia' + '/' + 'Instructions' + '/' + document_hash
            try:
                os.makedirs(document_folder)
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_content = document_response.content
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_content)
                with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                    document_text = PDFToTextService().text_from_pdf_path(
                        document_folder + '/' + self.language_code + '.pdf')
                    f.write(document_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': year,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return existed_docs

    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Estonia Annual Reports ===========================")
        existed_docs = []
        page_url = 'https://aastaraamat.aki.ee/'
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print('page not exist')
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        # get the newest annual report
        region_content = results_soup.find('div', class_='region region-content')
        block_core = region_content.find('div', class_='block-core')
        document_title = block_core.get_text().strip()
        print('document_title: ', document_title)
        document_hash = hashlib.md5(document_title.encode()).hexdigest()
        year = document_title.split()[-1].strip()
        block_views = region_content.find('div', class_='block-views')
        view_content = block_views.find('div', class_='view-content')
        for div in view_content.find_all('div', class_='views-row'):
            document = div.find('div', class_='views-field-title')
            title = document.get_text()
            document_href = document.find('a').get('href')
            if title != 'Aastaraamatu PDF':
                continue
            document_url = page_url + document_href
            document_source = self.get_source(page_url=document_url)
            if document_source is None:
                print('page not exist')
            document_soup = BeautifulSoup(document_source.text, 'html.parser')
            assert document_soup
            node_content = document_soup.find('div', class_='node__content')
            text_formatted = node_content.find('div', class_='text-formatted')
            article_href = text_formatted.find('a').get('href')
            article_url = page_url + article_href
            print('\turl: ', article_url)
            if document_hash in existing_docs and overwrite == False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash
            # document_folder = dpa_folder + '/estonia' + '/' + 'Annual Reports' + '/' + document_hash
            try:
                os.makedirs(document_folder)
                try:
                    document_response = requests.request('GET', article_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_content = document_response.content
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_content)
                with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                    document_text = PDFToTextService().text_from_pdf_path(
                        document_folder + '/' + self.language_code + '.pdf')
                    f.write(document_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': year,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)
            except FileExistsError:
                print("Directory path already exists, continue.")

        # older annual reports
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
        driver_doc.get(page_url)
        for i in range(1, 3):
            document = driver_doc.find_element_by_xpath('//*[@id="block-aastaraamat-main-menu"]/ul/li[3]/ul/li['+str(i)+']/a')
            document_href = document.get_attribute("href")
            page_source = self.get_source(page_url=document_href)
            if page_source is None:
                print('page not exist')
            document_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert document_soup
            region_content = document_soup.find('div', class_='region region-content')
            block_core = region_content.find('div', class_='block-core')
            document_title = block_core.get_text().strip()
            year_str = document_title.split()[-1].strip()
            # 2019 annual report
            if year_str == '2019':
                year = year_str
                print('document_title: ', document_title)
                print('\turl: ', document_href)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                block_views = region_content.find('div', class_='block-views')
                view_content = block_views.find('div', class_='view-content')
                for div in view_content.find_all('div', class_='views-row'):
                    document = div.find('div', class_='views-field-title')
                    title = document.get_text()
                    document_href = document.find('a').get('href')
                    if title != 'Aastaraamatu PDF':
                        continue
                    document_url = page_url + document_href
                    document_source = self.get_source(page_url=document_url)
                    if document_source is None:
                        print('page not exist')
                    document_soup = BeautifulSoup(document_source.text, 'html.parser')
                    assert document_soup
                    node_content = document_soup.find('div', class_='node__content')
                    text_formatted = node_content.find('div', class_='text-formatted')
                    article_href = text_formatted.find('a').get('href')
                    article_url = page_url + article_href
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
            # older than 2019
            else:
                block_system = region_content.find('div', class_='block-system')
                node_content = block_system.find('div', class_='node__content')
                for li in node_content.find('ul').find('li'):
                    article_title = li.get_text().strip()
                    document_title = article_title
                    year = article_title.split()[-1].strip()
                    if year < '2018':
                        continue
                    article_href = li.get('href')
                    document_hash = hashlib.md5(article_title.encode()).hexdigest()
                    print('document_title: ',  article_title)
                    article_url = page_url + article_href
                    print('\turl: ', article_url)
                    if document_hash in existing_docs and overwrite == False:
                        if to_print:
                            print('\tSkipping existing document:\t', document_hash)
                        continue
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash
            # document_folder = dpa_folder + '/estonia' + '/' + 'Annual Reports' + '/' + document_hash
            try:
                os.makedirs(document_folder)
                try:
                    document_response = requests.request('GET', article_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_content = document_response.content
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_content)
                with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                    document_text = PDFToTextService().text_from_pdf_path(
                        document_folder + '/' + self.language_code + '.pdf')
                    f.write(document_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                                self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': year,
                        'url': article_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)
            except FileExistsError:
                print("Directory path already exists, continue.")
        return existed_docs



import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import docx2txt
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
from selenium.webdriver.common.keys import Keys
import time

class Slovakia(DPA):
    def __init__(self, path=os.curdir):
        country_code='sk'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        source = {
            "host": "https://dataprotection.gov.sk",
            "start_path": "/uoou/sk/main-content/metodiky-uradu"
        }
        host = source['host']
        start_path = source['start_path']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        else:
            pager = page_soup.find('ul', class_='pager')
            if pager is not None:
                for li in pager.find_all('li', class_='pager-item'):
                    page_link = li.find('a')
                    if page_link is None:
                        continue
                    page_href = page_link.get('href')
                    pagination.add_item(host + page_href)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_fineAndReports(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_opinions(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_fineAndReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Slovakia Fine And Reports =========================")
        iteration = 1
        existed_docs = []
        source = {
            'host': 'https://dataprotection.gov.sk/',
            'start_path': '/uoou/sk/content/vyrocne-spravy',
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        page_source = self.get_source(page_url=page_url)
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        content_clearfix = results_soup.find('div', class_='content clearfix')
        # print(content_clearfix)
        assert content_clearfix

        for document in content_clearfix.find_all('div', class_='filefield-file'):
            valid_document = True
            document_href = document.find('a').get('href')
            document_title = document.find('a').get_text()
            # check the document date. Ignore all the documents released before 2018
            document_title_list = document_title.split()
            for i in range(len(document_title_list)):
                # ignore all the documents contain word which is digit && smaller than 2018
                if document_title_list[i].isdigit() and int(document_title_list[i]) < 2018:
                    valid_document = False
                    break
                # future check whether a word contain digit which is smaller than 2018
                sub_word = document_title_list[i].split('-')[0]
                if sub_word < '2018':
                    valid_document = False
                    break
            if valid_document:
                document_url = document_href

                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                print("\tdocument_url:\t", document_url)
                print("\tdocument_title:\t", document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                year_of_release = document_title_list[-1]
                print('\tyear: ', year_of_release)
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_content = document_response.content
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Fine & Reports' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        document_text = document_text.strip()
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': year_of_release,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return existed_docs

    def get_docs_opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Slovakia Opinions =========================")
        iteration = 1
        existed_docs = []
        pagination = self.update_pagination()
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            view_content_index = 1
            region_content = results_soup.find('div', class_='region-content')
            assert region_content
            content = region_content.find('div', class_='content')
            assert content
            # s1. Results
            for node_file in content.find_all('div', class_='node-file'):
                b = node_file.find('b')
                assert b
                date_str = b.get_text().split(' - ')[0]
                date_str = date_str.strip()
                tmp = datetime.datetime.strptime(date_str, '%d.%m.%Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                h2 = node_file.find('h2')
                assert h2
                result_link = h2.find('a')
                assert result_link
                # s2. Documents
                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                document_title = result_link.get_text()
                print('\tdocument_title:\t', document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://dataprotection.gov.sk"
                document_url = host + document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    document_soup = BeautifulSoup(document_response.text, 'html.parser')
                    assert document_soup
                    content_clearfix = document_soup.find('div', class_='content clearfix')
                    number = 0
                    for field_file in content_clearfix.find_all('div', class_='filefield-file'):
                        number += 1
                        assert len(field_file) > 0
                        article_url = field_file.find('a').get('href')
                        print('\tarticle_url: ', article_url)
                        file_number = ('_' + str(number)) if number > 0 else ''
                        # article is a docx file. Downloaded location "/documents/docx_files"
                        if article_url.endswith('.docx'):
                            # set a download path. Use selenium to download it
                            exec_path = WebdriverExecPolicy().get_system_path()
                            options = webdriver.ChromeOptions()
                            prefs = {"download.default_directory": document_folder}
                            options.add_argument('headless')
                            options.add_experimental_option("prefs", prefs)
                            driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                            # open google search, and put the document_url into search box, and press enter
                            driver_doc.get('https://www.google.com/')
                            box = driver_doc.find_element_by_xpath(  # google search box xpath
                                '/html/body/div[1]/div[3]/form/div[1]/div[1]/div[1]/div/div[2]/input')
                            box.send_keys(article_url)
                            box.send_keys(Keys.ENTER)
                            # the first result in the search is the file we need. Click the link to download it
                            download_section = driver_doc.find_element_by_class_name('tF2Cxc')
                            button = download_section.find_element_by_tag_name('h3').click()
                            time.sleep(5)
                            files = os.listdir(document_folder)
                            for index, file in enumerate(files):
                                os.rename(os.path.join(document_folder, file),
                                          os.path.join(document_folder, self.language_code + '.docx'))
                            document_content = docx2txt.process(document_folder + '/' + self.language_code + '.docx')
                            with open(
                                    document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                f.write(document_content)
                        else:
                            document_response = None
                            try:
                                document_response = requests.request('GET', article_url)
                                document_response.raise_for_status()
                            except requests.exceptions.HTTPError as error:
                                if to_print:
                                    print(error)
                                pass
                            if document_response is None:
                                continue
                            document_content = document_response.content
                            with open(document_folder + '/' + self.language_code + file_number + '.pdf', 'wb') as f:
                                f.write(document_content)

                            with open(document_folder + '/' + self.language_code + file_number + '.txt', 'wb') as f:
                                document_text = textract.process(document_folder + '/' + self.language_code + file_number + '.pdf')
                                document_text = document_text.strip()
                                f.write(document_text)

                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': document_title,
                                'md5': document_hash,
                                'releaseDate': date.strftime('%d/%m/%Y'),
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            # s0. Pagination
            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup)
            print('\n')
        return existed_docs







import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import time
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
import urllib3

class Poland(DPA):
    def __init__(self, path=os.curdir):
        country_code='po'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path="decisions"):
        source = {
            "host": "https://uodo.gov.pl",
            "start_path_decisions": "/pl/p/decyzje",
            "start_path_tutorials": "/pl/383"
        }
        host = source['host']
        if start_path == "decisions":
            start_path = source['start_path_decisions']
        else:
            start_path = source['start_path_tutorials']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        return pagination

    def get_source(self,  page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url, verify=False)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_News(existing_docs=[], overwrite=False, to_print=True)

        # below page have been stopped
        # added_docs += self.get_docs_Tutorials(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Poland Decision Documents =========================")
        iteration = 1
        existed_docs = []
        pagination = self.update_pagination()
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            decisions_container = results_soup.find('div', id='decisions-container')
            assert decisions_container
            # s1. Results
            for decision in decisions_container.find_all('a', class_='ui-decision'):
                time.sleep(2)
                date_div = decision.find('div', class_='ui-decision__date')
                assert date_div
                date_str = date_div.get_text()
                tmp = dateparser.parse(date_str, languages=[self.language_code])
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                result_link = decision.find('div', class_='ui-decision__title')
                assert result_link
                # s2. Documents
                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                document_title = result_link.get_text()
                print('\tdocument_title:\t', document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = decision.get('href')
                assert document_href
                document_url = document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                    print("\tDocument URL:\t", document_url)
                    print("\tdate:\t", date)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, verify=False)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                try:
                    document_soup = BeautifulSoup(document_response.text, 'html.parser')
                    article_content = document_soup.find('div', class_='decision-details')
                    document_text = article_content.get_text()
                    document_text = document_text.strip()
                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date.strftime('%d/%m/%Y'),
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        existed_docs.append(document_hash)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")
                except:
                    print('Document is not available')
        return existed_docs


    def get_docs_News(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Poland News =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        iteration = 1
        existed_docs = []
        source = {
             "host": "https://uodo.gov.pl",
            "start_path": "/pl/p/aktualnosci"
        }

        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        print('page_url:', page_url)

        page_source = self.get_source(page_url=page_url)
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert page_soup

        ui_card = page_soup.find('div', class_='ui-cards')
        assert ui_card
        for a in ui_card.find_all('a', class_='ui-card'):
            print('\n------------ Document: ' + str(iteration) + ' ------------')
            iteration += 1
            time.sleep(2)
            document_url = a.get('href')
            document_title = a.find('span', class_='ui-card__title').get_text()
            document_date = a.find('div', class_='ui-card__description').get_text().strip()
            print('\tDocument Title: ' + document_title)
            print('\tDocument URL: ' + document_url)
            print('\tDoument date: ', document_date)
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            print('\tdocument_hash: ', document_hash)
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:', document_hash)
                continue
            tmp = dateparser.parse(document_date)
            date = datetime.date(tmp.year, tmp.month, tmp.day)
            if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                continue
            try:
                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                driver_doc.get(document_url)

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'News' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    article = driver_doc.find_element_by_xpath('/html/body/div[1]/div/section/div[2]/div/div[2]/div[1]')

                    if article is None:
                        print('\tno document')
                        continue

                    document_text = article.text

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print('Directory path already exists, continue.')
            except:
                print('\tSomething went wrong getting the doc.')

        return existed_docs





    def get_docs_Tutorials(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Poland Tutorials =========================")
        iteration = 1

        existed_docs = []
        pagination = self.update_pagination(start_path="tutorials")
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            body_content = results_soup.find('div', class_='body-content')
            assert body_content
            # s1. Results
            for artLevel0 in body_content.find_all('div', class_='artLevel0'):
                time.sleep(2)
                result_link = artLevel0.find('a')
                assert result_link
                # s2. Documents
                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                document_title = result_link.get_text().strip()
                print('\tdocument_title:\t', document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                document_url = document_href
                print("\tdocument_url:\t", document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                driver_doc.get(document_url)
                article_content = driver_doc.find_element_by_id("article-content")
                document_text = article_content.text
                document_text = document_text.strip()
                date_div = driver_doc.find_elements_by_class_name('article-metric-button')
                date_str = ''
                for i in date_div:
                    date_str = i.text
                assert date_str
                tmp = dateparser.parse(date_str)
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                print("\tdate: ", date)
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Tutorials' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return existed_docs



import os
import shutil
import math
import requests
import json
import hashlib
import datetime
from pygdpr.models.dpa import DPA
import dateparser
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.models.common.pagination import Pagination
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy

class Croatia(DPA):
    def __init__(self, path=os.curdir):
        country_code='HR'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        source = {
            "host": "https://azop.hr",
            "start_path": "/novosti"
        }
        host = source['host']
        start_path = source['start_path']

        if pagination is None:
            page_url = host + start_path
            pagination = Pagination()
            pagination.add_item(page_url)
        else:
            # print("pagination was called.")
            pagination = Pagination()
            wp_pagenavi = page_soup.find('div', class_='wp-pagenavi')
            if wp_pagenavi is not None:
                for a in wp_pagenavi.find_all('a', class_='page'):
                    page_link = a.get('href')
                    # print("page_link was added:", page_link)
                    pagination.add_item(page_link)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver = webdriver.Chrome(options=options, executable_path=exec_path)
        driver.get(page_url)
        page_source = driver.page_source
        return page_source

    # Calls all scraper methods at once
    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        pagination = self.update_pagination()
        page_list = []
        print("\n========================= Croatia Decision Documents ===========================")
        #  s0. Pagination
        while pagination.has_next():
            page_number = ''
            page_url = pagination.get_next()
            for i in page_url.split('/'):
                if i.isdigit():
                    page_number = i
                    break
            # check if already download this page
            if page_number == '' and '1' in page_list:
                print("page already downloaded")
                continue
            if page_number in page_list:
                print("page already downloaded")
                continue
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source, 'html.parser')
            assert results_soup
            for post in results_soup.find_all('article', class_='post'):
                post_meta = post.find('p', class_='post-meta')
                assert post_meta
                published = post_meta.find('span', class_='published')
                date_str = published.get_text()
                tmp = dateparser.parse(date_str, languages=[self.language_code])
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                entry_title = post.find('h2', class_='entry-title')
                result_link = entry_title.find('a')
                assert result_link
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                print("document_title: ", document_title)
                print("\tDocument_hash:\t", document_hash)
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    document_href = result_link.get('href')
                    assert document_href
                    document_url = document_href

                    exec_path = WebdriverExecPolicy().get_system_path()
                    options = webdriver.ChromeOptions()
                    options.add_argument('headless')
                    driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                    driver_doc.get(document_url)
                    document_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
                    assert document_soup
                    et_pb_post_content = document_soup.find('div', class_='et_pb_post_content')
                    assert et_pb_post_content
                    document_text = et_pb_post_content.get_text()

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("Directory path already exists, continue.")
            wp_pagenavi = results_soup.find('div', class_='wp-pagenavi')
            pages = results_soup.find('span', class_='pages')
            current_page = pages.get_text().split()[1]
            page_list.append(current_page)
            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup)
        return added_docs


from lxml import html, etree
from lxml.etree import XPathEvalError
from pygdpr.models.common.reachability_node import Color
from pygdpr.specifications.reachable_node_specification import ReachableNodeSpecification
import os

class ReachabilityAnalysisService:
    def propagate_color(self, node):
        reachable_node = ReachableNodeSpecification()
        is_reachable = reachable_node.is_satisfied_by(node)
        if is_reachable:
            node.set_color(Color.BLACK)
        else:
            node.set_color(Color.WHITE)
        return node

    def analyze_gdpr_node(self, gdpr_node):
        gdpr_node.set_color(Color.GREY)
        for dpa_node in gdpr_node.get_children():
            self.analyze_dpa_node(dpa_node)
        gdpr_node = self.propagate_color(gdpr_node)
        return gdpr_node

    def analyze_dpa_node(self, dpa_node):
        dpa_node.set_color(Color.GREY)
        for source_node in dpa_node.get_children():
            self.analyze_source_node(source_node)
        dpa_node = self.propagate_color(dpa_node)
        return dpa_node

    def analyze_source_node(self, source_node):
        source_node.set_color(Color.GREY)
        page_source = source_node.source
        source_html = html.fromstring(page_source.encode())
        for label_node in source_node.get_children():
            self.analyze_label_node(source_html, label_node)
        source_node = self.propagate_color(source_node)
        return source_node

    def analyze_label_node(self, source_html, label_node):
        label_node.set_color(Color.GREY)
        for xpath_node in label_node.get_children():
            self.analyze_xpath_node(source_html, xpath_node)
        label_node = self.propagate_color(label_node)
        return label_node

    def analyze_xpath_node(self, source_html, xpath_node):
        xpath_node.set_color(Color.GREY)
        xpath = xpath_node.xpath
        try:
            elements = source_html.xpath(xpath)
            color = Color.BLACK if len(elements) > 0 else Color.WHITE
            xpath_node.set_color(color)
        except XPathEvalError:
            xpath_node.set_color(Color.WHITE)
        return xpath_node

    def perform_analysis(self, node):
        if node.get_color() == Color.BLACK:
            return node

        classname = node.__class__.__name__
        if classname == 'GDPRNode':
            return self.analyze_gdpr_node(node)
        elif classname == 'DPANode':
            return self.analyze_dpa_node(node)
        else:
            raise ValueError(f"The subclass {classname} is not supported, only GDPRNode and DPANode")
        return None



import os
import math
import requests
import json
import datetime
import hashlib
import textract
import urllib3
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import sys
import time
import re

class CzechRepublic(DPA):
    def __init__(self, path=os.curdir):
        country_code='CZ'
        super().__init__(country_code, path)

    # Added start_path as input parameter so different scraper methods can use update_pagination()
    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path=None):
        source = {
            "host": "https://www.uoou.cz",
            # "start_path": "/vismo/zobraz_dok.asp?id_ktg=901"
            "start_path": "/tiskove%2Dzpravy/ds-1017/p1=1017&tzv=1&pocet=25&stranka=1"
            # "start_path": "/na%2Daktualni%2Dtema/ds-1018/archiv=0&p1=1099&tzv=1&pocet=25&stranka=1"
        }
        host = source['host']
        start_path = start_path
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        else:
            strlistovani = page_soup.find('div', class_='strlistovani')
            if strlistovani is not None:
                for a in strlistovani.find_all('a'):
                    page_href = a.get('href')
                    pagination.add_item(host + page_href)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            # Set timeout param to low value
            # If IPv6 takes too long, the request will then switch to IPv4 quickly and use that protocol
            # This seems to be an issue for links with weaker infrastructure
            results_response = requests.request('GET', page_url, timeout=2)

            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    # Calls all scraper methods at once
    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_PressReleases(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Opinions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_CourtRulings(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_DecisionOfPresident(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_DecisionMakingActivites(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_CompletedInspections(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    # Always gets text from a document page
    # If there is a pdf link on the document page, downloads the pdf (and its text) as well
    # Does check dates
    def get_docs_PressReleases(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Czech Republic Press Releases ===========================")
        iteration = 1
        added_docs = []

        # We want to create the pagination object, then add the rest of the pages to visit to the
        # pagination object all at once, because calling update_pagination will insert all pages into the
        # pagination list each time.
        # We have to parse the first page before the while loop to do this
        pagination = self.update_pagination(start_path='/tiskove%2Dzpravy/ds-1017/p1=1017&tzv=1&pocet=25&stranka=1')
        initial_page_source = self.get_source(page_url='https://www.uoou.cz/tiskove%2Dzpravy/ds-1017/p1=1017&tzv=1&pocet=25&stranka=1')
        initial_results_soup = BeautifulSoup(initial_page_source.text, 'html.parser')
        pagination = self.update_pagination(pagination=pagination, page_soup=initial_results_soup)

        while pagination.has_next():
            page_url = pagination.get_next()
            print('\n------------ NEW PAGE ------------')
            if to_print:
                print('Page:\t', page_url)

            page_source = self.get_source(page_url=page_url)

            if page_source is None:
                print("Skipping page because page_source is None")
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            dok = results_soup.find('div', class_='obsah')
            assert dok
            ui = dok.find('ul', class_='ui')
            assert ui

            # Get the page number we are on -> if it greater than 3, skip
            # Only the first three pages contain documents made 2018 and after
            page_number = page_url[-1:]
            if int(page_number) > 3:
                print("\tSkipping page: " + page_number)
                continue

            for li in ui.find_all('li'):
                result_link = li.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                if document_href.startswith('http') is not True:
                    host = "https://www.uoou.cz"
                    document_url = host + document_href
                else:
                    document_url = document_href

                # TODO: One of the document_urls leads to a pdf link -> skip it for now
                if document_url == 'https://www.uoou.cz/assets/File.ashx?id_org=200144&id_dokumenty=31695':
                    continue

                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=2)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if document_response is None:
                    print("\tSkipping existing document: document_response is None")
                    continue

                # Created soup for the document link
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                # Obtain document date -> implemented simpler method that just checks the years
                div = li.find('div')
                if div is None:
                    print("\tSkipping existing document: div is None")
                    continue
                created_index = 0

                # Use this to get the date out
                div_text = div.get_text()

                m = re.search('(.+?) - ', div_text)
                if m:
                    found_date = m.group(1)
                else:
                    m = re.search('(.+?) – ', div_text)
                    if m:
                        found_date = m.group(1)

                found_date_year = found_date[-5:]

                if int(found_date_year):
                    found_date_year_int = int(found_date_year)
                    if found_date_year_int < 2018:
                        print("\tSkipping existing document: Document year is: " + found_date_year)
                        return added_docs
                    # Document year is 2018 or greater
                    else:
                        document_year = found_date_year
                # Can't convert the year string to an int for whatever reason -> just keep the doc anyways
                else:
                    document_year = "Date not available"

                # If significant pdf links exists, go to them and download
                obalcelek_tag = document_soup.find('div', id='obalcelek')
                if not obalcelek_tag is None:
                    a_tag = obalcelek_tag.find_all('a')
                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'PressReleases' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)
                        if a_tag:
                            for element in a_tag:
                                assert element
                                # Check if we can get a href and if that href contains the string 'File.ashx', which indicates
                                # the link is intended to be downloaded
                                if element.get('href') is not None and ('File.ashx' in element.get('href')):
                                    link_href = element.get('href')
                                    assert link_href
                                    link_url = 'https://www.uoou.cz' + link_href
                                    print("Link URL: " + link_url)

                                    link_response = None
                                    try:
                                        link_response = requests.request('GET', link_url, timeout=2)
                                        link_response.raise_for_status()
                                    except requests.exceptions.HTTPError as error:
                                        pass
                                    if link_response is None:
                                        continue

                                    # If get a link response, then write the contents of the file as a pdf and text
                                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                                        f.write(link_response.content)
                                    try:
                                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                                            link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                                            f.write(link_text)
                                    # If the link leads to a word document or a file format other than a pdf
                                    # -> skip text conversion
                                    except:
                                        pass

                                else:
                                    continue

                        # When we print document stuff, that means the document is not going to be thrown out
                        print('\n\t------------ Document: ' + str(iteration) + ' ------------')
                        iteration += 1
                        if to_print:
                            print("\tDocument:\t", document_hash)
                        obsah = document_soup.find('div', class_='obsah')
                        assert obsah
                        document_text = obsah.get_text()
                        document_text = document_text.lstrip()


                        with open(document_folder + '/' + self.language_code + 'Summary' + '.txt', 'w') as f:
                            f.write(document_text)
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': document_year,
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)

                        added_docs.append(document_hash)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")
                # Don't need to call update_pagination -> it already has all the pages
        return added_docs

    # Always gets text from a document page
    # If there is a pdf link on the document page, downloads the pdf (and its text) as well
    # Doesn't get dates -> html is to inconsistent -> just scrape first page of link instead (other pages
    # contains links that are too old)
    def get_docs_Opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Czech Republic Opinions ===========================")
        iteration = 1
        added_docs = []

        # We want to create the pagination object, then add the rest of the pages to visit to the
        # pagination object all at once, because calling update_pagination will insert all pages into the
        # pagination list each time.
        # We have to parse the first page before the while loop to do this
        pagination = self.update_pagination(start_path='/na%2Daktualni%2Dtema/ds-1018/archiv=0&p1=1099&tzv=1&pocet=25&stranka=1')
        initial_page_source = self.get_source(page_url='https://www.uoou.cz/na%2Daktualni%2Dtema/ds-1018/archiv=0&p1=1099&tzv=1&pocet=25&stranka=1')
        initial_results_soup = BeautifulSoup(initial_page_source.text, 'html.parser')
        pagination = self.update_pagination(pagination=pagination, page_soup=initial_results_soup)

        while pagination.has_next():
            page_url = pagination.get_next()
            print('\n------------ NEW PAGE ------------')
            if to_print:
                print('\tPage:\t', page_url)

            page_source = self.get_source(page_url=page_url)

            if page_source is None:
                print("Skipping page because page_source is None")
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            dok = results_soup.find('div', class_='obsah')
            assert dok
            ui = dok.find('ul', class_='ui')
            assert ui

            # Get the page number we are on -> if it greater than 1, skip
            # Only the first page contains documents made 2018 and after
            page_number = page_url[-1:]
            if int(page_number) > 1:
                print("\tSkipping page: " + page_number)
                continue

            for li in ui.find_all('li'):
                result_link = li.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                if document_href.startswith('http') is not True:
                    host = "https://www.uoou.cz"
                    document_url = host + document_href
                else:
                    document_url = document_href

                # TODO: One of the document_urls leads to a pdf link -> skip it for now
                if document_url == 'https://www.uoou.cz/assets/File.ashx?id_org=200144&id_dokumenty=31695':
                    continue

                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=2)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if document_response is None:
                    print("\tSkipping existing document: document_response is None")
                    continue

                # Created soup for the document link
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                # If significant pdf links exists, go to them and download
                obalcelek_tag = document_soup.find('div', id='obalcelek')
                if not obalcelek_tag is None:
                    a_tag = obalcelek_tag.find_all('a')
                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)
                        if a_tag:
                            for element in a_tag:
                                assert element
                                # Check if we can get a href and if that href contains the string 'File.ashx', which indicates
                                # the link is intended to be downloaded
                                if element.get('href') is not None and ('File.ashx' in element.get('href')) :
                                    link_href = element.get('href')
                                    assert link_href
                                    link_url = 'https://www.uoou.cz' + link_href
                                    print("Link URL: " + link_url)

                                    link_response = None
                                    try:
                                        link_response = requests.request('GET', link_url, timeout=2)
                                        link_response.raise_for_status()
                                    except requests.exceptions.HTTPError as error:
                                        pass
                                    if link_response is None:
                                        continue

                                    # If get a link reponse, then write the contents of the file as a pdf and text
                                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                                        f.write(link_response.content)
                                    try:
                                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                                            link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                                            f.write(link_text)
                                    # If the link leads to a word document or a file format other than a pdf
                                    # -> skip text conversion
                                    except:
                                        pass
                                else:
                                    continue

                        # Don't try to find the document date -> website html is too inconsistent
                        # When we print document stuff, that means the document is not going to be thrown out
                        print('\n\t------------ Document: ' + str(iteration) + ' ------------')
                        iteration += 1
                        if to_print:
                            print("\tDocument:\t", document_hash)
                        obsah = document_soup.find('div', class_='obsah')
                        assert obsah
                        document_text = obsah.get_text()
                        document_text = document_text.lstrip()
                        with open(document_folder + '/' + self.language_code + 'Summary' + '.txt', 'w') as f:
                            f.write(document_text)
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': "Date not available",
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        added_docs.append(document_hash)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")
                # Don't need to call update_pagination -> it already has all the pages
        return added_docs

    # Does date checking
    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Czech Republic Annual Reports ===========================")
        added_docs = []

        page_url = 'https://www.uoou.cz/vyrocni-zprava/ds-2089/archiv=0&p1=2087'
        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        dok = results_soup.find('div', id='obalcelek')
        assert dok
        ui = dok.find('ul', class_='ui')
        assert ui

        iteration = 1
        for li in ui.find_all('li'):
            result_link = li.find('a')

            document_title = result_link.get_text()
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite == False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = result_link.get('href')

            assert document_href

            if document_href.startswith('http') is not True:
                host = "https://www.uoou.cz"
                document_url = host + document_href
            else:
                document_url = document_href
            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=2)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                pass
            if document_response is None:
                print("\tSkipping existing document: document_response is None")
                continue
            # If document year is earlier than 2018, skip it
            document_year = result_link.get_text()[-4:]
            if int(document_year) < 2018:
                print("\tSkipping existing document: Document year is: " + document_year)
                return added_docs

            # When we print document stuff, that means the document is not going to be thrown out
            print('\n\t------------ Document: ' + str(iteration) + ' ------------')
            iteration += 1
            if to_print:
                print("\tDocument:\t", document_hash)

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'AnnualReports' + '/' + document_hash
            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_response.content)

                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                    f.write(link_text)

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_year,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)

                print('------------------------')
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return added_docs

    # Does date checking to an extent -> if date is obtainable check it, if not, then include doc anyways
    def get_docs_CourtRulings(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Czech Republic Court Rulings ===========================")
        added_docs = []
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        page_url = 'https://www.uoou.cz/ceska-republika/ds-2850/archiv=0&p1=1271'
        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        kategorie_tag = results_soup.find('div', id='kategorie')
        assert kategorie_tag
        ui_tag = kategorie_tag.find('ul', class_='ui')
        assert ui_tag

        iteration = 1
        for li in ui_tag.find_all('li'):
            result_link = li.find('a')

            document_href = result_link.get('href')
            assert document_href

            if document_href.startswith('http') is not True:
                host = "https://www.uoou.cz"
                document_url = host + document_href
            else:
                document_url = document_href

            print("---VISITING LINK: " + document_url + ' ---')

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=4, verify=False)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                pass
            if document_response is None:
                print("\tSkipping existing link: document_response is None")
                continue

            # Created soup for the document link
            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            link_kategorie_tag = document_soup.find('div', id='kategorie')
            assert link_kategorie_tag
            link_ui_tag = link_kategorie_tag.find('ul', class_='ui')
            assert link_ui_tag

            for link_li in link_ui_tag.find_all('li'):
                doc_tag = link_li.find('a')
                assert doc_tag

                # Get title and make hash for document
                document_title = doc_tag.get_text()

                # Try to obtain the document year from the last 4 characters of document_title
                # Not all titles contain the date here, but a lot do, so its worth the check
                document_year = document_title[-4:]

                if document_year.isnumeric():
                    if int(document_year) < 2018:
                        print("\tSkipping document because it is out of date")
                        continue
                    else:
                        pass
                else:
                    document_year = "Document year not available"

                # Ensure the document year is a string
                document_year = str(document_year)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                link_href = doc_tag.get('href')
                assert link_href

                if link_href.startswith('http') is not True:
                    link_host = "https://www.uoou.cz"
                    link_url = link_host + link_href
                else:
                    link_url = link_href

                link_response = None
                try:
                    link_response = requests.request('GET', link_url, timeout=10, verify=False)
                    link_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if link_response is None:
                    print("\tSkipping existing document: link_response is None")
                    continue
                # link_response now contains the pdf contents

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'CourtRulings' + '/' + document_hash

                # The conditionals check if the link is either a pdf or a text doc. This is done by identifying
                # strings within the url to indicate the document type. Note that the conditional checks may excluded
                # extremely dated docs. Very old links use different identifiers for pdf and text links than newer ones.

                # Document isn't a pdf
                if 'GetText.aspx' in link_url:

                    # When we print document stuff, that means the document is not going to be thrown out
                    print('\n\t------------ (TEXT) Document: ' + str(iteration) + ' ------------')
                    print("\tLink URL: " + link_url)
                    iteration += 1
                    if to_print:
                        print("\tDocument:\t", document_hash)

                    # Create soup for the link
                    link_soup = BeautifulSoup(link_response.text, 'html.parser')
                    assert link_soup
                    body = link_soup.find('form', id='WordForm')
                    assert body

                    body_text = body.get_text()
                    assert body_text

                    try:
                        os.makedirs(document_folder)

                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            f.write(body_text.encode())

                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': "Date not implemented yet",
                                'url': link_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")

                # Document is a pdf
                # Some of the links end with a space character at the end
                elif link_url.endswith('.pdf') or link_url.endswith('.pdf ') or 'File.ashx' in link_url:
                    # When we print document stuff, that means the document is not going to be thrown out
                    print('\n\t------------ (PDF) Document: ' + str(iteration) + ' ------------')
                    link_url = link_url.strip()
                    print("\tLink URL: " + link_url)
                    iteration += 1
                    if to_print:
                        print("\tDocument:\t", document_hash)

                    try:
                        os.makedirs(document_folder)
                        with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                            f.write(link_response.content)
                        time.sleep(5)

                        # TODO: Fix bug where converting pdf to text fails occasionally (could be request header issue)
                        try:
                            with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                link_text = PDFToTextService().text_from_pdf_path(document_folder + '/' + self.language_code + '.pdf')
                                f.write(link_text)
                        except:
                            print("\tFailed to convert pdf to text")
                            pass

                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': "Date not implemented yet",
                                'url': link_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")
                # If link doesn't lead to a pdf or text page, scrap it
                else:
                    print('\tSkipping document link because it does not appear to lead to a text page/pdf.'
                          '\n\t Or document link may be so outdated that text/pdf check fails')
                    continue

                added_docs.append(document_hash)
                print('\t------------------------')

        return added_docs

    # Always gets text from a document page
    # If there is a pdf link on the document page, downloads the pdf (and its text) as well
    # Checks document year, only visits first page for now since rest of pages only contain
    # outdated documents
    def get_docs_DecisionMakingActivites(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Czech Republic Decision Making Activites ===========================")
        iteration = 1
        added_docs = []

        # We want to create the pagination object, then add the rest of the pages to visit to the
        # pagination object all at once, because calling update_pagination will insert all pages into the
        # pagination list each time.
        # We have to parse the first page before the while loop to do this
        pagination = self.update_pagination(
            start_path='/z-rozhodovaci-cinnosti-uradu/ds-1022/p1=1277&tzv=1&pocet=25&stranka=1')
        initial_page_source = self.get_source(
            page_url='https://www.uoou.cz/z-rozhodovaci-cinnosti-uradu/ds-1022/p1=1277&tzv=1&pocet=25&stranka=1')
        initial_results_soup = BeautifulSoup(initial_page_source.text, 'html.parser')
        pagination = self.update_pagination(pagination=pagination, page_soup=initial_results_soup)

        while pagination.has_next():
            page_url = pagination.get_next()
            print('\n------------ NEW PAGE ------------')
            if to_print:
                print('\tPage:\t', page_url)

            page_source = self.get_source(page_url=page_url)

            if page_source is None:
                print("Skipping page because page_source is None")
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            dok = results_soup.find('div', class_='obsah')
            assert dok
            ui = dok.find('ul', class_='ui')
            assert ui

            # Get the page number we are on -> if it greater than 1, skip
            # Only the first page contains documents made 2018 and after
            page_number = page_url[-1:]
            if int(page_number) > 1:
                print("\tSkipping page: " + page_number)
                continue

            for li in ui.find_all('li'):
                result_link = li.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                if document_href.startswith('http') is not True:
                    host = "https://www.uoou.cz"
                    document_url = host + document_href
                else:
                    document_url = document_href

                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=2)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if document_response is None:
                    print("\tSkipping existing document: document_response is None")
                    continue

                # Created soup for the document link
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                # Check document date
                popis = document_soup.find('div', class_='popis dpopis')
                assert popis
                date_string = popis.get_text()
                document_year = date_string[-4:]
                # Document was made before 2018, skip it
                if int(document_year) < 2018:
                    print("\tSkipping outdated document existing document:\t", document_hash)
                    continue

                # If significant pdf links exists, go to them and download
                obalcelek_tag = document_soup.find('div', id='obalcelek')
                if obalcelek_tag is not None:
                    a_tag = obalcelek_tag.find_all('a')
                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'Decision-Making Activities' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)
                        if a_tag:
                            for element in a_tag:
                                assert element
                                # Check if we can get a href and if that href contains the string 'File.ashx', which indicates
                                # the link is intended to be downloaded
                                if element.get('href') is not None and ('File.ashx' in element.get('href')):
                                    link_href = element.get('href')
                                    assert link_href
                                    link_url = 'https://www.uoou.cz' + link_href
                                    print("Link URL: " + link_url)

                                    link_response = None
                                    try:
                                        link_response = requests.request('GET', link_url, timeout=2)
                                        link_response.raise_for_status()
                                    except requests.exceptions.HTTPError as error:
                                        pass
                                    if link_response is None:
                                        continue

                                    # If get a link reponse, then write the contents of the file as a pdf and text
                                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                                        f.write(link_response.content)
                                    try:
                                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                                            link_text = textract.process(
                                                document_folder + '/' + self.language_code + '.pdf')
                                            f.write(link_text)
                                    # If the link leads to a word document or a file format other than a pdf
                                    # -> skip text conversion
                                    except:
                                        pass
                                else:
                                    continue

                        # Don't try to find the document date -> website html is too inconsistent
                        # When we print document stuff, that means the document is not going to be thrown out
                        print('\n\t------------ Document: ' + str(iteration) + ' ------------')
                        iteration += 1
                        if to_print:
                            print("\tDocument:\t", document_hash)
                        print('\t------------------------')
                        obsah = document_soup.find('div', class_='obsah')
                        assert obsah
                        document_text = obsah.get_text()
                        document_text = document_text.lstrip()
                        with open(document_folder + '/' + self.language_code + 'Summary' + '.txt', 'w') as f:
                            f.write(document_text)
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': document_year,
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        added_docs.append(document_hash)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")
                # Don't need to call update_pagination -> it already has all the pages
        return added_docs

    # No date checking
    def get_docs_DecisionOfPresident(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Czech Republic Decision Of President ===========================")
        added_docs = []

        page_url = 'https://www.uoou.cz/rozhodnuti-predsedy-uradu/ds-3815/archiv=0&p1=1277'
        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        dok = results_soup.find('div', id='obalcelek')
        assert dok
        obsah = dok.find('div', class_='obsah')
        assert obsah

        iteration = 1
        for ul in obsah.find_all('ul'):
            assert ul
            for li in ul.find_all('li'):
                assert li

                result_link = li.find('a')

                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = result_link.get('href')
                assert document_href

                if document_href.startswith('http') is not True:
                    host = "https://www.uoou.cz"
                    document_url = host + document_href
                else:
                    document_url = document_href

                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if document_response is None:
                    print("\tSkipping existing document: document_response is None")
                    continue

                # When we print document stuff, that means the document is not going to be thrown out
                print('\n\t------------ Document: ' + str(iteration) + ' ------------')
                iteration += 1
                if to_print:
                    print("\tDocument:\t", document_hash)

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Decision of President' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_response.content)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(link_text)

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title.strip()
                            },
                            'md5': document_hash,
                            'releaseDate': 'date not available',
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)

                    print('\t------------------------')
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

        return added_docs

    # THE REMAINDER OF THIS FILE IS DEALING WITH THE COMPLETED INSPECTIONS LINK

    # This method is designed to be called by parent method get_docs_CompletedInspections() that
    # scrapes completed inspections.
    # Doesn't get date of specific text documents
    # argument: page_url -> the page we want to scrape. This needs to be a link to a yearly control activities page,
    # where we can see the headings for Finances, Marketing, Education, etc...
    # argument: folder_title -> this is the name of the folder we will store everything in
    def get_docs_DecisionChecksControlActivites(self, existing_docs=[], overwrite=False, to_print=True, page_url=None,
                                                folder_title=None):
        added_docs = []

        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        dok = results_soup.find('div', id='obalcelek')
        assert dok
        ui = dok.find('ul', class_='ui')
        assert ui

        iteration = 1
        for li in ui.find_all('li'):
            assert li
            result_link = li.find('a')
            assert result_link

            document_href = result_link.get('href')
            assert document_href

            if document_href.startswith('http') is not True:
                host = "https://www.uoou.cz"
                document_url = host + document_href
            else:
                document_url = document_href

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=2)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                pass
            if document_response is None:
                print("\tSkipping outer link: document_response is None")
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            # Now get the text from each of the links on this second page
            obalcelek = document_soup.find('div', id='obalcelek')
            assert obalcelek
            document_ui = obalcelek.find('ul', class_='ui')
            assert document_ui

            for document_li in document_ui.find_all('li'):
                assert document_li
                text_link = document_li.find('a')
                assert text_link

                document_title = text_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document - hash already exists:\t', document_hash)
                    continue

                text_href = text_link.get('href')
                assert text_href

                if text_href.startswith('http') is not True:
                    host = "https://www.uoou.cz"
                    text_url = host + text_href
                else:
                    text_url = text_href

                text_response = None
                try:
                    text_response = requests.request('GET', text_url, timeout=2)
                    text_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if text_response is None:
                    print("\tSkipping existing document: document_response is None")
                    continue

                text_soup = BeautifulSoup(text_response.text, 'html.parser')
                assert text_soup

                text_obalcelek = text_soup.find('div', id='obalcelek')
                assert text_obalcelek
                text_body = text_obalcelek.find('div', id='stred')
                assert text_body

                # When we print document info, that means the document is not going to be thrown out
                print('\n\t------------ Document: ' + str(iteration) + ' ------------')
                print('\tDocument title: ' + document_title)
                iteration += 1
                if to_print:
                    print("\tDocument:\t", document_hash)

                # Now store the text in the appropriate folder
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Inspections' + '/' + folder_title + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        f.write(text_body.get_text().encode())

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': 'date not available',
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)

                    print('\t------------------------')
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

        return added_docs

    # Same method as get_docs_DecisionChecksControlActivities, except this handles older link version where
    # scraper only needs to go 1 layer deep to access text documents
    def get_docs_DecisionChecksControlActivites2018Below(self, existing_docs=[], overwrite=False, to_print=True, page_url=None,
                                                folder_title=None):

        added_docs = []

        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        dok = results_soup.find('div', id='obalcelek')
        assert dok
        ui = dok.find('ul', class_='ui')
        assert ui

        iteration = 1
        for li in ui.find_all('li'):
            assert li
            result_link = li.find('a')
            assert result_link

            document_title = result_link.get_text()
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document - hash already exists:\t', document_hash)
                continue

            document_href = result_link.get('href')
            assert document_href

            if document_href.startswith('http') is not True:
                host = "https://www.uoou.cz"
                document_url = host + document_href
            else:
                document_url = document_href

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=2)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                pass
            if document_response is None:
                print("\tSkipping outer link: document_response is None")
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            text_obalcelek = document_soup.find('div', id='obalcelek')
            assert text_obalcelek
            text_body = text_obalcelek.find('div', id='stred')
            assert text_body

            # When we print document info, that means the document is not going to be thrown out
            print('\n\t------------ Document: ' + str(iteration) + ' ------------')
            print('\tDocument title: ' + document_title)
            iteration += 1
            if to_print:
                print("\tDocument:\t", document_hash)

            # Now store the text in the appropriate folder
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Inspections' + '/' + folder_title + '/' + document_hash
            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    f.write(text_body.get_text().encode())

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': 'date not available',
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
                print('\t------------------------')
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs

    # This method is designed to be called by parent method get_docs_CompletedInspections() that
    # scrapes completed inspections.
    # Doesn't get date of specific text documents
    # argument: page_url -> the page we want to scrape. This needs to be a link to a yearly unsolicited commercial
    # communications page, where we can see the headings for company inspections
    # argument: folder_title -> this is the name of the folder we will store everything in
    def get_docs_DecisionChecksUnsolicitedCommerical(self, existing_docs=[], overwrite=False, to_print=True, page_url=None,
                                                folder_title=None):
        added_docs = []

        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        dok = results_soup.find('div', id='obalcelek')
        assert dok
        ui = dok.find('ul', class_='ui')
        assert ui

        iteration = 1
        hash_iteration = 1
        for li in ui.find_all('li'):
            assert li
            result_link = li.find('a')
            assert result_link

            # The titles the page gives are often the same, we add iteration number to title to distinguish
            document_title = result_link.get_text() + ' ' + str(hash_iteration)
            hash_iteration += 1
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document - hash already exists:\t', document_hash)
                continue

            document_href = result_link.get('href')
            assert document_href

            if document_href.startswith('http') is not True:
                host = "https://www.uoou.cz"
                document_url = host + document_href
            else:
                document_url = document_href

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=2)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                pass
            if document_response is None:
                print("\tSkipping outer link: document_response is None")
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            text_obalcelek = document_soup.find('div', id='obalcelek')
            assert text_obalcelek
            text_body = text_obalcelek.find('div', id='stred')
            assert text_body

            # When we print document info, that means the document is not going to be thrown out
            print('\n\t------------ Document: ' + str(iteration) + ' ------------')
            print('\tDocument title: ' + document_title)
            iteration += 1
            if to_print:
                print("\tDocument:\t", document_hash)

            # Now store the text in the appropriate folder
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Inspections' + '/' + folder_title + '/' + document_hash
            try:
                os.makedirs(document_folder)
                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    f.write(text_body.get_text().encode())

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': 'date not available',
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
                print('\t------------------------')
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs

    # This is the method to be called when scraping Completed Inspections documents (part of Decisions).
    # This method starts at the main Completed Inspections page and visits each "Checks for XXXX" link.
    # If year for the link is less than 2018, it is not scraped.
    # Once inside "Check for XXXX" link, the method examines the links on the new page and determines the
    # appropriate scraper methods to call, depending on if links lead to Control Activities or Commercial
    # Communications type documents.
    def get_docs_CompletedInspections(self, existing_docs=[], overwrite=False, to_print=True):

        print("\n========================= Czech Republic Completed Inspections ===========================")
        added_docs = []

        page_url = 'https://www.uoou.cz/ukoncene-kontroly/ds-5649/archiv=0&p1=1277'
        if to_print:
            print('\tPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            print("Skipping page because page_source is None")
            sys.exit()

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        dok = results_soup.find('div', id='obalcelek')
        assert dok
        ui = dok.find('ul', class_='ui')
        assert ui

        for li in ui.find_all('li'):
            assert li
            result_link = li.find('a')
            assert result_link

            # The titles the page gives are often the same, we add iteration number to title to distinguish
            document_title = result_link.get_text()
            inspection_date = document_title[-4:]
            if int(inspection_date) < 2018:
                print('\tSkipping inspections for year: ' + inspection_date)
                continue

            document_href = result_link.get('href')
            assert document_href

            if document_href.startswith('http') is not True:
                host = "https://www.uoou.cz"
                document_url = host + document_href
            else:
                document_url = document_href

            print('URL for Checks: ' + inspection_date + ' is ' + document_url)

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=2)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                pass
            if document_response is None:
                print("\tSkipping inspection link: document_response is None")
                continue

            soup = BeautifulSoup(document_response.text, 'html.parser')
            assert soup

            # Now look at control actives and unsolicited commercial communications
            obalcelek = soup.find('div', id='obalcelek')
            assert obalcelek
            page_ui = obalcelek.find('ul', class_='ui')
            assert page_ui

            for page_li in page_ui.find_all('li'):
                assert page_li

                next_link = page_li.find('a')
                assert next_link

                link_title = next_link.get_text()
                assert link_title

                link_href = next_link.get('href')
                assert link_href

                if link_href.startswith('http') is not True:
                    host = "https://www.uoou.cz"
                    link_url = host + link_href
                else:
                    link_url = link_href

                # If links leads to a control actives pages and the year is 2018 or lower
                if 'kontrolni' in link_url and int(inspection_date) <= 2018:
                    # If this is control activities for second half of the year
                    if '2.' in link_title:
                        added_docs += self.get_docs_DecisionChecksControlActivites2018Below(page_url=link_url,
                                        folder_title=(inspection_date + ' Control Activities - 2nd half of the year'))

                    elif '1.' in link_title:
                        added_docs += self.get_docs_DecisionChecksControlActivites2018Below(page_url=link_url,
                                        folder_title=(inspection_date + ' Control Activities - 1st half of the year'))

                    else:
                        print('Failed to scrape Control Activities link')

                # If links leads to a regular control actives pages
                elif 'kontrolni' in link_url:
                    # If this is control activities for second half of the year
                    if '2.' in link_title:
                        added_docs += self.get_docs_DecisionChecksControlActivites(page_url=link_url,
                                        folder_title=(inspection_date + ' Control Activities - 2nd half of the year'))

                    elif '1.' in link_title:
                        added_docs += self.get_docs_DecisionChecksControlActivites(page_url=link_url,
                                        folder_title=(inspection_date + ' Control Activities - 1st half of the year'))

                    else:
                        print('Failed to scrape Control Activities link')

                # Link leads to unsolicited commercial communications page
                elif 'nevyzadana' in link_url:
                    # If this is unsolicited commercial communications for second half of the year
                    if '2.' in link_title:
                        added_docs += self.get_docs_DecisionChecksUnsolicitedCommerical(page_url=link_url,
                                        folder_title=(inspection_date
                                                    + ' Unsolicited Commercial Communications - 2nd half of the year'))

                    elif '1.' in link_title:
                        added_docs += self.get_docs_DecisionChecksUnsolicitedCommerical(page_url=link_url,
                                        folder_title=(inspection_date
                                                    + ' Unsolicited Commercial Communications - 1st half of the year'))
                    else:
                        print('Failed to scrape Unsolicited Commercial Communications link')

                else:
                    print("Couldn't determine type for link: " + link_url)

        return added_docs



import os
import math
import requests
import json
import datetime
import hashlib

import textract

from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy

class Finland(DPA):
    def __init__(self, path=os.curdir):
        country_code='FI'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path=None):
        page_url = 'https://tietosuoja.fi/en/current-issues'
        pagination = Pagination()
        pages = page_soup.find('ul', class_='nav-pills')
        # print('pages: ', pages)
        for li in pages.find_all('li', class_='nav-item'):
            a = li.find('a')
            page_url = a.get('href')
            pagination.add_item(page_url)
        return pagination


    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:

            results_response = requests.request('GET', page_url, timeout=2)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    # some docs are written is english, and some are in Finnish
    # for the english version docs, change the language.code as 'en'
    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):

        page_url = 'https://tietosuoja.fi/en/current-issues'
        page_source = self.get_source(page_url)
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        pagination = self.update_pagination(page_soup=page_soup)

        print("\n========================= Finland Documents ===========================")
        added_docs = []


        if to_print:
            print('\nPAGE:\t', page_url)

        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)

            iterator = 1
            page_source = self.get_source(page_url)
            page_soup = BeautifulSoup(page_source.text, 'html.parser')
            ul = page_soup.find('ul', class_='results')
            for li in ul.find_all('li', class_='list__item'):
                span_date = li.find('span', class_='date')
                assert span_date
                date_str = span_date.get_text()
                tmp = datetime.datetime.strptime(date_str, '%d.%m.%Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)

                if date.year < 2018:
                    print("Skipping outdated document")
                    return added_docs

                ul_year = date.year
                result_link = li.find('a')
                assert result_link
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                print('\n------------ Document: ' + str(iterator) + ' ------------')
                print('\tDocument Title: ' + document_title)
                print("\tdate: ", date)
                iterator += 1

                document_href = result_link.get('href')
                assert document_href
                host = "https://tietosuoja.fi"
                document_url = host + document_href

                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                # Document parse object
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                # Obtain document text
                news_page = document_soup.find('div', class_='news-page')
                if news_page is None:
                    continue
                document_text = news_page.get_text()
                document_text = document_text.lstrip()

                # Look at all links contained in the document
                document_tags = news_page.find_all(['p', 'ul'])
                assert document_tags

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + str(ul_year) + ' Finland Documents' + '/' + document_hash
                # document_folder = dpa_folder + '/finland' + '/' + str(ul_year) + ' Finland Documents' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    # Use this when naming the files for pdf links we download. Only increment if use it in a name.
                    link_iterator = 1
                    # This second iterator is for if we have finlex links on the page too
                    # -> don't want to confuse with pdf stuff
                    link_iterator_finlex_links = 1
                    for tags in document_tags:
                        if tags.find('a') is not None:
                            a_tag = tags.find_all('a')
                            for i in range(len(a_tag)):
                                href = a_tag[i].get('href')
                                assert href

                                if href.startswith('http'):
                                    url = href
                                else:
                                    url = "https://tietosuoja.fi" + href

                                if '.pdf' in url:
                                    pdf_response = None
                                    try:
                                        pdf_response = requests.request('GET', url)
                                        pdf_response.raise_for_status()
                                    except requests.exceptions.HTTPError as error:
                                        if to_print:
                                            print(error)
                                        pass
                                    if pdf_response is None:
                                        continue

                                    print("\tDownloading PDF: " + url)
                                    # Write pdf and its text to files

                                    with open(document_folder + '/' + 'en' + str(link_iterator) + '.pdf', 'wb') as f:
                                        f.write(pdf_response.content)
                                    with open(document_folder + '/' + 'en' + str(link_iterator) + '.txt', 'wb') as f:
                                        link_text = textract.process(document_folder + '/' + 'en' + str(link_iterator) + '.pdf')
                                        f.write(link_text)

                                    link_iterator += 1

                                elif url.startswith('https://finlex') or url.startswith('https://www.finlex'):
                                    text_response = None
                                    try:
                                        text_response = requests.request('GET', url)
                                        text_response.raise_for_status()
                                    except requests.exceptions.HTTPError as error:
                                        if to_print:
                                            print(error)
                                        pass
                                    if text_response is None:
                                        continue

                                    print("\tDownloading Finlex text: " + url)
                                    text_soup = BeautifulSoup(text_response.text, 'html.parser')
                                    assert text_soup

                                    body = text_soup.find('div', id='main-content')
                                    # print(body)
                                    assert body
                                    body_text = body.get_text()
                                    body_text = body_text.lstrip()
                                    assert body_text

                                    with open(document_folder + '/' + self.language_code + 'Finlex' + str(link_iterator_finlex_links) + '.txt', 'wb') as f:
                                        f.write(body_text.encode())
                                    link_iterator_finlex_links += 1

                                # Link is not useful
                                else:
                                    continue

                        # The <p tag doesn't provide a link
                        else:
                            continue

                    with open(document_folder + '/' + 'en' + 'Summary' + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return added_docs



import os
import math
import socket

import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
import sys
import urllib3.exceptions

from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
import urllib3

class Slovenia(DPA):
    def __init__(self, path=os.curdir):
        country_code='SL'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        source = {
            "host": "https://www.ip-rs.si",
            "start_path": "/mnenja-gdpr/"
        }
        host = source['host']
        start_path = source['start_path']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        elif page_soup is not None:
            nav = page_soup.find('div', class_='page pr')
            assert nav

            article = nav.find('article', class_='c9')
            assert article

            advsea_next = article.find('span', class_='advsea-next')
            assert advsea_next

            a_tag = advsea_next.find('a')
            assert a_tag

            href = a_tag.get('href')
            next_url = 'https://www.ip-rs.si/' + href
            #print("\nAdding page: " + next_url + " to pagination object")
            pagination.add_item(next_url)
        else:
            print("update_pagination() added zero links")

        return pagination

    def get_source(self, page_url=None, driver=None):
        assert page_url is not None
        results_response = None
        try:
            results_response = requests.request('GET', page_url, verify=False)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Opinions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Reports(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guidelines(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Slovenia Opinions =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        added_docs = []
        pagination = self.update_pagination()

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)

            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            nav = results_soup.find('div', class_='page pr')
            assert nav

            article = nav.find('article', class_='c9')
            assert article

            advsea = article.find('div', class_='advsea-results-list')
            assert advsea

            for tr in advsea.find_all('tr')[1:]:
                assert tr

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1

                td_date = tr.find('td', align='center')
                assert td_date

                document_date = td_date.get_text()
                assert document_date

                print('\tDocument Date: ' + document_date)

                tmp = datetime.datetime.strptime(document_date, '%d.%m.%Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)

                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    print('Remaining documents are out of date')
                    return added_docs

                td = tr.find('td', align='left')
                assert td

                a = td.find('a')
                assert a

                href = a.get('href')
                assert href

                # THE HREF'S IN THE HTML ARE MISSING THE BEGINNING '/', SO IT MUST BE ADDED HERE
                document_url = 'https://www.ip-rs.si/' + href
                assert document_url

                print("\tDocument Link: " + document_url)

                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()

                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                except requests.exceptions.ConnectionError as error:
                    if to_print:
                        print(error)
                    pass
                except urllib3.exceptions.MaxRetryError as error:
                    if to_print:
                        print(error)
                    pass
                except urllib3.exceptions.NewConnectionError as error:
                    if to_print:
                        print(error)
                    pass
                except socket.gaierror as error:
                    if to_print:
                        print(error)
                    pass

                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                article = document_soup.find('article', class_='c9')
                assert article

                h1 = article.find('header').find('h1')
                assert h1

                document_title = h1.get_text()
                print('\tDocument Title: ' + document_title)

                # Get the document text
                article_text = article.get_text()

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(article_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup)
        return added_docs

    def get_docs_Guidelines(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Slovenia Guidelines =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        added_docs = []

        page_url = 'https://www.ip-rs.si/publikacije/priro%C4%8Dniki-in-smernice/'
        if to_print:
            print('\nPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        nav = results_soup.find('div', class_='page pr')
        assert nav

        article = nav.find('article', class_='c9')
        assert article

        table = article.find('table', class_='contenttable')
        assert table

        tr = table.find('tr', class_='holder')
        assert tr

        iteration = 1
        for td in tr.find_all('td'):
            assert td

            a_tag = td.find_all('a')[1]
            assert a_tag

            print('\n------------ Document ' + str(iteration) + ' ------------')
            iteration += 1

            document_title = a_tag.get_text()
            print('\tDocument Title: ' + document_title)

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = a_tag.get('href')
            assert document_href

            document_url = 'https://www.ip-rs.si/' + document_href

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            # Now download the pdf
            pdf_article = document_soup.find('article', class_='c9')
            assert pdf_article

            # Get the document date
            table = pdf_article.find('table', class_='contenttable')
            assert table

            tr_table = table.find_all('tr')
            assert tr_table

            date_tr = tr_table[4]
            assert date_tr

            date_td = date_tr.find_all('td')[1]
            assert date_td

            date_p = date_td.find('p')

            document_date = date_p.get_text()
            print('\tDocument Date: ' + document_date)

            tmp = datetime.datetime.strptime(document_date, '%d.%m.%Y')
            date = datetime.date(tmp.year, tmp.month, tmp.day)

            if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                print('Before GDPR adopted')
                continue

            # Skip old reports
            if int(document_date[-4:]) < 2018:
                print('Skipping outdated document')
                return added_docs

            # Find the pdf link
            pdf_p = pdf_article.find('p', align='center')
            assert pdf_p

            pdf_a = pdf_p.find('a')
            assert pdf_a

            pdf_href = pdf_a.get('href')
            assert pdf_href

            # Sometimes the href doesn't start with a '/'
            if pdf_href.startswith('/'):
                pdf_url = 'https://www.ip-rs.si' + pdf_href
            else:
                pdf_url = 'https://www.ip-rs.si/' + pdf_href

            print("\tPDF Link: " + pdf_url)

            pdf_response = None
            try:
                pdf_response = requests.request('GET', pdf_url)
                pdf_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if pdf_response is None:
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Guidelines' '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(pdf_response.content)
                try:
                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        document_text = PDFToTextService().text_from_pdf_path(document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                except:
                    print('Failed to convert PDF to text')
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_date,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return added_docs

    def get_docs_Reports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Slovenia Reports =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        added_docs = []

        page_url = 'https://www.ip-rs.si/publikacije/letna-poro%C4%8Dila/'
        if to_print:
            print('\nPage:\t', page_url)

        page_source = self.get_source(page_url=page_url)

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        nav = results_soup.find('div', class_='page pr')
        assert nav

        article = nav.find('article', class_='c9')
        assert article

        table = article.find('table', class_='contenttable')
        assert table

        tr = table.find('tr', class_='holder')
        assert tr

        iteration = 1
        for td in tr.find_all('td'):
            assert td

            a_tag = td.find_all('a')[1]
            assert a_tag

            print('\n------------ Document ' + str(iteration) + ' ------------')
            iteration += 1

            document_title = a_tag.get_text()
            print('\tDocument Title: ' + document_title)

            document_date = document_title[-4:]
            print('\tDocument Date: ' + document_date)

            # Skip old reports
            if int(document_date[-4:]) < 2018:
                print('Skipping outdated document')
                return added_docs

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = a_tag.get('href')
            assert document_href

            if document_href.startswith('http'):
                document_url = document_href
            else:
                document_url = 'https://www.ip-rs.si/' + document_href
                print('\tDocument URL: ' + document_url)

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Reports' '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_response.content)

                try:
                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        document_text = PDFToTextService().text_from_pdf_path(document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                except:
                    print('Failed to convert PDF to text')

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_date,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs
